import Foundation
#if os(macOS)
import Cocoa
#elseif os(iOS)
import UIKit
#endif
import PlaygroundSupport

#if os(macOS)
typealias SystemFarbe = NSColor
typealias View = NSView
#elseif os(iOS)
typealias SystemFarbe = UIColor
typealias View = UIView
#endif

/**
 Zufallsgenerator auf der Basis der UNIX random Funktion
 */
class Random
{

	/**
     Konstruktor legt den Startwert fest für eine wiederholbare Folge
     */
    init ()
    {
    	srandom(0)
    }
    
    /**
     Zufällige Kommazahl aus dem Bereich [0; 1[
     - returns: Zufallszahl aus 0; 1[
     */
    func ZufallKomma () -> Float
    {
    	return Float (arc4random()) / Float (UINT32_MAX)
    }

    /**
     Zufällige ganze Zahl aus dem Bereich [0; top[
     - parameters:
     	- top: Obergrenze des Ergebnisbereichs
     - returns: ganze Zufallszahl aus 0; top[
     */
    func ZufallGanz (top: Int) -> Int
    {
    	return Int (ZufallKomma() * Float (top))
    }
}

/**
 interne Klasse für die Roboterdaten, die aus einer Karolwelt-Datei
 ausgelesen werden.
 */
struct RoboDat
{
    var posX: Int = 0
    var posY: Int = 0
    var direct: Character
    
    /**
    Initialisiert die Roboterdaten mit Position und Blickrichtung.
    - parameters:
    	- x: x-Position
    	- y: y-Position
        - b: Blickrichtung
    */
    init (x: Int, y: Int, b: Character)
    {
        posX = x
        posY = y
        direct = b
    }
}

/**
 Interner Enum für die Farben von  Ziegeln und Marken.
 */
enum BildFarben
{
	case keine, rot, gelb, grün, blau, schwarz
    
    static func FarbeFürString (farbe: String) -> BildFarben
    {
        switch farbe.lowercased()
        {
            case "rot":
                return rot
            case "gelb":
                return gelb
            case "grün", "gruen":
                return grün
            case "blau":
                return blau
            case "schwarz":
                return schwarz
            default:
                return keine
        }
    }
}

/**
Ausnahmen die von Karol geworfen werden können
*/
enum RoboterFehler: Error
{
	/** Karol kann einen Schritt wegen Hindernissen nicht ausführen */
    case bewegungNichtMoeglich
    /** Karol kann einen Schritt wegen Höhenunterschied nicht ausführen */
    case sprungHoehe
    /** Hinlegen eines Ziegels nicht möglich */
    case hinlegenNichtMoeglich
    /** Aufheben eines Ziegels nicht möglich */
    case aufhebenNichtMoeglich
    /** Aufstellen eines Quaders nicht möglich */
    case aufstellenNichtMoeglich
    /** Abbauen eines Quaders nicht möglich */
    case abbauenNichtMoeglich
}

/**
Ausnahmen die von der Welt geworfen werden können
*/
enum WeltFehler: Error
{
	/** Datei enthält keine Karol-Welt */
    case keineKarolWelt
    /** Die Welt wurde verändert, reload nicht möglich */
	case karolWeltVerändert
	/** Datei nicht gefunden */
    case weltNichtGefunden
}

#if os(macOS)
/**
Hilfsklasse für die Open- / Savedialoge
*/
class OpenSavePanels: NSObject, NSOpenSavePanelDelegate
{
    private let manager = FileManager.default
    
	override init()
    {
        super.init()
    }

    func panel(_ sender: Any, shouldEnable url: URL) -> Bool
    {
        do
        {
            let teile = url.path.lowercased().components(separatedBy: ".")
            let res = try manager.attributesOfItem(atPath: url.path)
            return (res[FileAttributeKey.type] as! NSString) as FileAttributeType == FileAttributeType.typeDirectory ||
            	(teile.count > 1 && (teile [teile.count-1] == "kdw" || teile [teile.count-1] == "jkw"))
        }
        catch
        {
            return false
        }
    }

    func DateinameOeffenGeben () -> String?
    {
        let panel = NSOpenPanel()
        panel.delegate = self
        
        let res = panel.runModal()
        panel.orderOut(self)
        NSApp.abortModal()
        if res == NSApplication.ModalResponse.OK
        {
        	let url = panel.urls[0]
            return url.path
        }
        return nil
    }

    func DateinamenSicherungGeben (alterName: String) -> String?
    {
        let panel = NSSavePanel()
        panel.showsTagField = false
        panel.nameFieldStringValue = alterName == "" ? "Unbekannt.jkw" : alterName
        let res = panel.runModal()
        panel.orderOut(self)
        NSApp.abortModal()
        if res == NSApplication.ModalResponse.OK
        {
        	let url = panel.url
            return (url != nil) ? url!.path : nil
        }
        else
        {
        	return nil
        }
    }
}
#endif

/**
 Realisierung einer Karolwelt in Java
 
 RobotKarol ist eine Programmiersprache, die für Schülerinnen und Schüler
 zum Erlernen des Programmierens und zur Einführung in die Algorithmik gedacht ist.
 Für die Entwicklung von RobotKarol-Programmen steht eine eigene Programmierumgebung
 zur Verfügung (www.schule.bayern.de/karol).
 JavaKarol ist eine Weiterführung dieses Konzepts bei dem die Objekte einer Karolwelt
 mit Java erzeugt und bearbeitet werden können.
 
 Eine Karolwelt ist ein rechteckiger, dreidimensionaler Raum, dessen Boden mit einem
 Quadratmuster ausgelegt ist und der von vier Wänden umgeben ist. Die Karolwelt kann
 verschiedene Objekte wie Ziegel, Quader und Marken enthalten. In dieser Welt
 können sich mehrere Objekte der Klasse Roboter bewegen. (Z.Zt. maximal 9)
 
 Die Klasse Welt ist eine Umsetzung einer Karolwelt in die Programmiersprache Java.
 Eine Welt kann auch aufgrund einer Karolwelt-Datei erstellt werden, wobei
 JavaKarol auch die Weltdefinitionen von RobotKarol verwenden kann. Geeignete
 Welten können somit mit der Programmierumgebung von RobotKarol generiert werden.
 
 Methoden der Welt:
 Speichern(); BildSpeichern(); ZurueckSetzen(); ZiegelVerstreuen();
 
 - author: Ulli Freiberger
 
 - version: 1.0 (9.1.2008)

 Umsetzung der Klassen nach Swift für die Entwicklung mit Xcode.
 
 - author: Albert Wiedemann
 - version: 1.0
 - version: 2.0 (03.02.2016) Farben für Ziegel und Marken
 - version: 3.0 (23.09.2019) Änderngen JavaKarol 3.0
 */
open class Welt
{
	var weltFenster: WeltFenster! = nil
	
	let maxHoehe = 31
	let st_Nichts: Int8  = 0
	let st_Ziegel: Int8  = 1
	let st_Quader: Int8  = 2
	let st_Roboter: Int8 = 3
	
	var areaBreite: Int = 0, areaLaenge: Int = 0, areaHoehe: Int = 0  // Weltkoordinatensystem x: 1..areaBreite; y: 1..areaLaenge
	var areaMarkiert: [[(ist: Bool, farbe: BildFarben)]] = [[(ist: Bool, farbe: BildFarben)]] () // true, wenn an dieser Stelle eine Markierung ist

	var areaStapelInhalt: [[[(was: Int8, farbe: BildFarben)]]] = [[[(was: Int8, farbe: BildFarben)]]] ()		    // Inhalt des Stapels je Höhe, kann nichts oder Ziegel oder Quader sein; unterster liegt auf 0, Farbe für Ziegel oder keine für Quader)

	static let maxRoboter = 9						    // z. Zt. maximal 9 Roboter
	var nextRoboter: Int = 1					// nächste zu vergebende Roboterkennung
	var alleRoboter: [Roboter] = [Roboter] ()	// alle in der Welt angemeldeten Roboter class Roboter
	var geladeneRoboter: [RoboDat] = [RoboDat] ()	// alle Robotereinstellungen aus einer geladenen Karolwelt class RoboDat
	
	var weltDateiname: String = ""	            // Name der Karolwelt-Datei die Grundlage beim Erzeugen des Objekts war; "" bei leerer Welt
    var ist2d: Bool = false
    
    #if os(macOS)
    var panele: OpenSavePanels = OpenSavePanels()
    #endif

    /**
    Erzeugen einer neuen, leeren Welt in der angegebenen Größe

    Es muss immer erst eine Welt angelegt werden.
    Die Welt öffnet automatisch das zugehörige Anzeigefenster.
    - parameters:
        - breite: Breite der Welt 1..breite; entspricht x-Koordinate
        - laenge: Länge (Tiefe) der Welt 1..laenge ; entspricht y-Koordinate
        - hoehe: Höhe der Welt 1..31
    */
	public init (breite: Int, laenge: Int, hoehe: Int)
	{
        WeltLeeren(breite: breite, laenge: laenge, hoehe: hoehe)
		initView()
	}

    /**
    Setzt die Welt auf den Anangszustand zurück
    - parameters:
        - breite: Breite der Welt 1..breite; entspricht x-Koordinate
        - laenge: Länge (Tiefe) der Welt 1..laenge ; entspricht y-Koordinate
        - hoehe: Höhe der Welt 1..31
    */
    private func WeltLeeren (breite: Int, laenge: Int, hoehe: Int)
    {
        areaBreite = max(breite,1)
		areaLaenge = max(laenge,1)
		areaHoehe  = max(min(hoehe,maxHoehe),1)
		// alle Markierungen löschen
        areaMarkiert.removeAll()
        for _ in 0...areaBreite
        {
        	areaMarkiert.append ([(ist: Bool, farbe: BildFarben)] (repeating: (false, BildFarben.keine), count: areaLaenge+1))
        }
		// alle Stapelinhalte auf nichts setzen
        areaStapelInhalt.removeAll()
        for i in 0...areaBreite
        {
        	areaStapelInhalt.append ([[(was: Int8, farbe: BildFarben)]] ())
            for _ in 0...areaLaenge
            {
            	areaStapelInhalt [i].append([(was: Int8, farbe: BildFarben)] ())
            }
        }
		
		alleRoboter.removeAll()
		geladeneRoboter.removeAll()
    }
    
    /**
    Erzeugen einer neuen Welt aus einer Karolwelt-Datei
    mit einem vorgegeben Dateinamen.

    Die Einstellungen der Welt und der darin enthaltenen
    Ziegel, Quader und Markierungen werden aus der Karolwelt-Datei
    geladen und die entsprechenden Objekte erzeugt.
    Es sind sowohl Karolwelten von RobotKarol als auch von JavaKarol
    möglich.
    Die Welt öffnet automatisch das zugehörige Anzeigefenster.
    - parameters:
        - weltDatei: Dateiname der Karolwelt-Datei
          Falls die Datei nicht vorhanden ist oder sich nicht im
          Klassenpfad befindet wird ein Dateiauswahlfenster geöffnet.
    */
    public init (weltDatei: String)
    {
        do
        {
            try loadKarolFile(absFileName: validWorldFile(fileName: weltDatei),reload: false)
            initView()
        }
        catch
        {
        }
    }

    
	/**
	 Erzeugen einer neuen Welt aus einer Karolwelt-Datei
	 ohne Angabe eines Dateinamens.
	 
	 Es wird immer erst ein Dateiauswahlfenster geöffnet, in dem
	 die Karolwelt-Datei ausgewählt werden kann.
	 Die Einstellungen der Welt und der darin enthaltenen
	 Ziegel, Quader und Markierungen werden aus dieser Karolwelt-Datei
	 geladen und die entsprechenden Objekte erzeugt.
	 Es sind sowohl Karolwelten von RobotKarol als auch von JavaKarol
	 möglich.
	 Die Welt öffnet automatisch das zugehörige Anzeigefenster.
	 */
    public init ()
    {
        #if os(macOS)
        let datei = panele.DateinameOeffenGeben()
        #else
        let datei: String? = nil
        #endif
        assert(datei != nil, "Keine Karolwelt-Datei ausgewählt.")
        do
        {
            try loadKarolFile(absFileName: validWorldFile(fileName: datei! as String),reload: false)
            initView()
        }
        catch
        {
        }
    }
	
    /**
    Die Einstellungen der Welt zusammen mit den Positionen der
    Ziegel, Quader, Markierungen und Robotern wird gespeichert.
     
    Das Dateiformat ist stets jkw (JavaKarolWelt). Dieses Dateiformat ist
    eine Erweiterung des Dateiformats kwd (KarolWeltDeutsch) von RobotKarol.
     
    - parameters:
        - dateiname: Name der Karolwelt-Datei die erstellt werden soll.
    */
	open func Speichern(dateiname: String)
	{
		saveKarolFile(fileName: dateiname)
	}
	
	
    /**
    Die Welt und die darin enthaltenen Grafikobjekte werden als Bild
    gespeichert.
    Mögliche Dateiformate sind bmp und jpg
    AW: Mögliche Dateiformate ergänzt, möglich sind bmp, jpg, png, gif, tiff und pdf
     
    - parameters:
        - dateiname: Name der Bilddatei die erstellt werden soll.
    */
	open func BildSpeichern(dateiname: String)
	{
		saveWorldImage(fileName: dateiname);
	}

	
	/**
    Die Welt und die darin enthaltenen Objekte werden wieder in den
    Startzustand (wie bei der Instanzierung) zurücksetzen.
    */
	open func ZurueckSetzen()
	{
		reset()
		weltFenster.zeichnen()
		weltFenster.fehlerAusgabeLoeschen()
	}
	
	/**
    Wechselt zwischen 2- und 3-dimensionaler Darstellung des Roboters
    - parameters:
    	- ist: 2-dimensional, wenn wahr
    */
    open func ZweiDSetzen (ist: Bool)
    {
    	ist2d = ist
		weltFenster.zeichnen();
    }

    /**
    Eine vorgegebene Anzahl an Ziegel wird zufällig über die Welt
    verstreut, wobei eine maximale Stapelhöhe eingehalten wird.
    - parameters:
        - anzahlZiegel: Anzahl an Ziegel, die verstreut werden
        - maxStapelhoehe: maximale Höhe, die dabei ein Stapel annehmen kann
    */
	open func ZiegelVerstreuen(anzahlZiegel: Int, maxStapelhoehe: Int)
	{
        var x, y, anz, maxB, i, j: Int;
        let zufall =  Random()
        var f: BildFarben;

        anz = min (anzahlZiegel, areaBreite*areaLaenge*areaHoehe)
        maxB = min (max (maxStapelhoehe, 1), areaHoehe)
        
        j = 0
        i = 0
        while (i < anz) && (j < 3 * anz)
        {
            x = zufall.ZufallGanz(top: areaBreite) + 1
            y = zufall.ZufallGanz(top: areaLaenge) + 1
            if (areaStapelInhalt[x][y].count < maxB) && (areaStapelInhalt[x][y].count == 0  || areaStapelInhalt[x][y][0].was == st_Ziegel)
            {
                switch zufall.ZufallGanz(top: 4)
                {
                    case 0:
                        f = BildFarben.rot
                    case 1:
                        f = BildFarben.gelb
                    case 2:
                        f = BildFarben.grün
                    default:
                        f = BildFarben.blau
                }
                push (x: x, y: y, what: st_Ziegel, f: f)
            }
            else
            {
                i -= 1
            }
            j += 1
            i += 1
        }
		weltFenster.zeichnen();
	}
	
	
	/*
	 * interne Methoden
	 */

	private func initView()
	{
		weltFenster = WeltFenster(frame: CGRect(x: 0.0, y: 0.0, width: 450, height: 375))
        weltFenster.WeltSetzen(w: self)
    }	

	private func clear()
	{
		
		// alle Markierungen löschen
		for j in 0 ..< areaMarkiert.count
        {
        	for i in 0 ..< areaMarkiert [j].count
            {
            	areaMarkiert [j] [i] = (false, BildFarben.keine)
            }
        }
		// alle Stapelinhalte auf nichts setzen
		for j in 0 ..< areaStapelInhalt.count
        {
        	for i in 0 ..< areaStapelInhalt [j].count
            {
            	areaStapelInhalt [j] [i].removeAll(keepingCapacity: true)
            }
        }
	}
	
    /*
     * Testet auf gültige Namenskonvention und lesbare Datei,
     * Opendialog wird nicht nochmals aufgerufen.
     */
	private func validWorldFile(fileName: String) throws -> String
	{
        let res: String = fileName
        let teile = fileName.lowercased().components(separatedBy: ".")
        guard (teile.count > 1) && (teile [teile.count-1] == "kdw" || teile [teile.count-1] == "jkw") else
        {
            throw WeltFehler.keineKarolWelt
        }
        return res
	}
	
	private func loadKarolFile(absFileName: String, reload: Bool) throws
	{
        let url = CFBundleCopyResourceURL(CFBundleGetMainBundle (), absFileName as CFString, nil, nil)
        guard let d = try? FileHandle(forReadingFrom: url! as URL).readDataToEndOfFile() else
        {
        	throw WeltFehler.weltNichtGefunden
        }
        let tokens = NSString(bytes: (d as NSData).bytes, length: d.count, encoding: String.Encoding.utf8.rawValue)?.components(separatedBy: " ")
        
        var pos: Int
        var breite, laenge, hoehe, posx, posy: Int
        var direct: Character
        
        let version = tokens![0]

        pos = 1
        if version == "KarolVersion1Deutsch" || version == "KarolVersion2Deutsch" || version == "JavaKarolVersion1" || version == "JavaKarolVersion1.1" || version == "KarolVersion3.0" || version == "JavaKarolVersion3.0"
        {
        	breite = Int(tokens![pos])!
            pos += 1
        	laenge = Int(tokens![pos])!
            pos += 1
        	hoehe = Int(tokens![pos])!
            pos += 1
            if reload
            {
            	if breite != areaBreite || laenge != areaLaenge || hoehe != areaHoehe
                {
                	throw WeltFehler.karolWeltVerändert
                }
            }
            else
            {
            	WeltLeeren(breite: breite, laenge: laenge, hoehe: hoehe)
                weltDateiname = absFileName
            }
            geladeneRoboter.removeAll()
        	posx = Int(tokens![pos])! + 1
            pos += 1
        	posy = Int(tokens![pos])! + 1
            pos += 1
        	direct = ["S", "W", "N", "O"] [Int(tokens![pos])!]
            pos += 1
            geladeneRoboter.append(RoboDat(x: posx, y: posy, b: direct))
            
            for a in 1...areaBreite
            {
            	for b in 1...areaLaenge
                {
                	for _ in 0..<areaHoehe
                    {
                    	let token = tokens![pos]
                        pos += 1
                        if token == "z"
                        {
                            push(x: a, y: b, what: st_Ziegel, f: BildFarben.rot)
                        }
                        else if token == "A"
                        {
                            push(x: a, y: b, what: st_Ziegel, f: BildFarben.rot)
                        }
                        else if token == "B"
                        {
                            push(x: a, y: b, what: st_Ziegel, f: BildFarben.gelb)
                        }
                        else if token == "C"
                        {
                            push(x: a, y: b, what: st_Ziegel, f: BildFarben.blau)
                        }
                        else if token == "D"
                        {
                            push(x: a, y: b, what: st_Ziegel, f: BildFarben.grün)
                        }
                        else if token == "q"
                        {
                        	push(x: a, y: b, what: st_Quader, f: BildFarben.keine)
                        }
                    }
                    let token = tokens![pos]
                    pos += 1
                    if token == "m"
                    {
                        setMarker(x: a, y: b, f: BildFarben.gelb)
                    }
                    else if token == "K"
                    {
                        setMarker(x: a, y: b, f: BildFarben.rot)
                    }
                    else if token == "L"
                    {
                        setMarker(x: a, y: b, f: BildFarben.gelb)
                    }
                    else if token == "M"
                    {
                        setMarker(x: a, y: b, f: BildFarben.blau)
                    }
                    else if token == "N"
                    {
                        setMarker(x: a, y: b, f: BildFarben.grün)
                    }
                    else if token == "O"
                    {
                        setMarker(x: a, y: b, f: BildFarben.schwarz)
                    }
                }
            }
            
            if version == "JavaKarolVersion1" || version == "JavaKarolVersion1.1" || version == "JavaKarolVersion3.0"
            {
                geladeneRoboter.removeAll()
                while pos < tokens!.count - 2
                {
                    posx = Int(tokens![pos])! + 1
                    pos += 1
                    posy = Int(tokens![pos])! + 1
                    pos += 1
                    direct = tokens![pos][tokens![pos].startIndex]
                    pos += 1
                    geladeneRoboter.append(RoboDat(x: posx, y: posy, b: direct))
                }
            }
        }
        else
        {
        	throw WeltFehler.keineKarolWelt
        }
    }

	private func saveKarolFile(fileName: String)
	{
        let m = FileManager.default
        if !m.fileExists(atPath: fileName)
        {
        	m.createFile(atPath: fileName, contents: nil, attributes: [:])
        }
        if let f = FileHandle(forWritingAtPath: fileName)
        {
            f.truncateFile(atOffset: 0)
            var res: String
            res = "JavaKarolVersion3.0 \(areaBreite) \(areaLaenge) \(areaHoehe) "
            if alleRoboter.count > 0
            {
                let robo = alleRoboter[0]
                res += "\(robo.PositionXGeben()-1) \(robo.PositionYGeben()-1) \(["S":"0", "W":"1", "N":"2", "O":"3"] [robo.BlickrichtungGeben()]!) "
            }
            else
            {
                res += "0 0 0 "
            }
            for a in 1 ... areaBreite
            {
                for b in 1 ... areaLaenge
                {
                    for c in 0 ..< areaHoehe
                    {
                        switch getPart(x: a, y: b, z: c)
                        {
                            case (st_Nichts, _):
                                res += "n "
                            case (st_Ziegel, BildFarben.rot):
                                res += "A "
                            case (st_Ziegel, BildFarben.gelb):
                                res += "B "
                            case (st_Ziegel, BildFarben.blau):
                                res += "C "
                            case (st_Ziegel, BildFarben.grün):
                                res += "D "
                            case (st_Quader, _):
                                res += "q "
                            default:
                                res += "x "
                                assertionFailure("Unbekanntes Weltelement beim Speichern")
                        }
                    }
                    switch getMarker(x: a, y: b)
                    {
                        case (true, BildFarben.rot):
                             res += "K "
                        case (true, BildFarben.gelb):
                             res += "L "
                        case (true, BildFarben.blau):
                             res += "M "
                        case (true, BildFarben.grün):
                             res += "N "
                        case (true, BildFarben.schwarz):
                             res += "O "
                        default:
                        	res += "o "
                    }
                }
            }
            for robo in alleRoboter
            {
                res += "\(robo.PositionXGeben()-1) \(robo.PositionYGeben()-1) \(robo.BlickrichtungGeben()) "
            }
            f.write(res.data(using: String.Encoding.utf8)!)
            weltDateiname = fileName
        }
        else
        {
        	assert(false, "Datei kann nicht geschrieben werden: \(fileName)")
        }
    }

	private func saveWorldImage(fileName: String)
	{
		#if os(macOS)
        var ok: Bool = false
        var bittype: NSBitmapImageRep.FileType = NSBitmapImageRep.FileType.png
        var typ: String = ""
        var data: Data?
        let teile = fileName.components(separatedBy: ".")
        if teile.count > 1
        {
        	typ = teile [teile.count - 1].lowercased()
            if typ == "png"
            {
            	ok = true
            }
            else if typ == "jpg" || typ == "jepg"
            {
            	bittype = NSBitmapImageRep.FileType.jpeg
                ok = true
            }
            else if typ == "tif" || typ == "tiff"
            {
            	bittype = NSBitmapImageRep.FileType.tiff
                ok = true
            }
            else if typ == "bmp"
            {
            	bittype = NSBitmapImageRep.FileType.bmp
                ok = true
            }
            else if typ == "gif"
            {
            	bittype = NSBitmapImageRep.FileType.gif
                ok = true
            }
            else if typ == "pdf"
            {
                ok = true
            }
        }
		if !ok
        {
        	assert(false, "Unbekannter Dateityp: \(fileName)")
            return
        }

        let view = weltFenster.weltAnzeige!
        let size = view.bounds.size
        if typ == "pdf"
        {
        	data = view.dataWithPDF(inside: view.bounds)
        }
        else
        {
            let image = NSImage(size: NSSize(width: size.width + 10, height: size.height + 20))
            image.lockFocus()
            view.draw(view.bounds)
            //let rep = NSBitmapImageRep(focusedViewRect: NSRect(x: 0, y: 0, width: size.width + 10, height: size.height + 20))
            image.unlockFocus()
            let rep = view.bitmapImageRepForCachingDisplay(in: NSRect(x: 0, y: 0, width: size.width + 10, height: size.height + 20))!
            view.cacheDisplay(in: NSRect(x: 0, y: 0, width: size.width + 10, height: size.height + 20), to: rep)
            data = rep.representation(using: bittype, properties: [:])
        }

        let m = FileManager.default
        if !m.fileExists(atPath: fileName)
        {
        	m.createFile(atPath: fileName, contents: nil, attributes: [:])
        }
        if let f = FileHandle(forWritingAtPath: fileName)
        {
            f.truncateFile(atOffset: 0)
            f.write(data!)
        }
        else
        {
        	assert(false, "Datei kann nicht geschrieben werden: \(fileName)")
        }
        #else
        var ok: Bool = false
        var typ: String = ""
        var data: Data?
        let teile = fileName.components(separatedBy: ".")
        if teile.count > 1
        {
            typ = teile [teile.count - 1].lowercased()
            if typ == "png"
            {
                ok = true
            }
            else if typ == "jpg" || typ == "jepg"
            {
                ok = true
            }
            /*%%%%%
            else if typ == "pdf"
            {
                ok = true
            }
            */
        }
        if !ok
        {
            assert(false, "Unbekannter Dateityp: \(fileName)")
            return
        }
        let view = weltFenster.weltAnzeige!
        let size = view.bounds.size
		UIGraphicsBeginImageContext(size)
    	view.draw(view.bounds)
    	let img = UIGraphicsGetImageFromCurrentImageContext()
    	UIGraphicsEndImageContext()
    	if typ == "png"
    	{
        	data = img?.pngData()
        }
        else
        {
        	data = img?.jpegData(compressionQuality: 9.5)
        }

        let m = FileManager.default
        if !m.fileExists(atPath: fileName)
        {
            m.createFile(atPath: fileName, contents: nil, attributes: [:])
        }
        if let f = FileHandle(forWritingAtPath: fileName)
        {
            f.truncateFile(atOffset: 0)
            f.write(data!)
        }
        else
        {
            assert(false, "Datei kann nicht geschrieben werden: \(fileName)")
        }
		#endif
    }

	private func reset()
	{
		clear()
		if !weltDateiname.isEmpty
        {
            do
            {
                try loadKarolFile(absFileName: weltDateiname, reload: true)
            }
            catch
            {
            }
        }
		for r in alleRoboter
		{
			r.reset()
		}
	}


	/*
	 * Methoden für die Verbindung zu den Robotern
	 */
	
    /**
    Roboter bei der Welt anmelden
    - parameters:
        - anmeldeRoboter: der anzumeldende Roboter
        - neu: wenn wahr, wird eine neue Roboterkennung vergeben, sonst die aktuelle verwendet
    - returns: Kennung des angemeldeteten Roboters
    */
    func roboterAnmelden(anmeldeRoboter: Roboter, neu: Bool) -> Int
    {
		var neueKennung, posX, posY: Int
		
		if neu
		{
            assert(nextRoboter <= Welt.maxRoboter, "Maximal \(Welt.maxRoboter) Roboter erlaubt.")
			neueKennung = nextRoboter
		}
		else
        {
			neueKennung = anmeldeRoboter.KennungGeben()
        }
		
		posX = anmeldeRoboter.PositionXGeben()
		posY = anmeldeRoboter.PositionYGeben()
			
        assert(getRobotID(x: posX, y: posY) <= 0, "An dieser Stelle steht schon ein Roboter.")
        assert(!isStone(x: posX, y: posY), "An dieser Stelle steht ein Quader.")
			
		if neu
        {
			nextRoboter += 1
        }
		alleRoboter.append(anmeldeRoboter)
		return neueKennung
    }

	/**
    Meldet den Roboter bei der Welt ab
    - parameters:
    	- abmeldeRoboter der abzumeldende Roboter
    */
    func roboterAbmelden(abmeldeRoboter: Roboter)
    {
		var i: Int
        i = -1;
        for (i1, r) in  alleRoboter.enumerated()
        {
        	if r === abmeldeRoboter
            {
            	i = i1
            }
        }
        alleRoboter.remove (at: i)
    }
	
	
	/**
    Holt Grunddaten für einen neuen Roboter
    wenn vorhanden, dann aus der KarolWelt-Datei
    sonst Defaultbelegung
    - returns: Roboterdaten
    */
	func roboterDatenAbholen() -> RoboDat
	{
		var anzRoboter, anzRoboDaten: Int
		var robodat = RoboDat(x: 1,y: 1,b: "S")
		
		anzRoboter   = alleRoboter.count
		anzRoboDaten = geladeneRoboter.count
		
		if (anzRoboter < anzRoboDaten)
		{
			robodat = geladeneRoboter [anzRoboter]
		}
		return robodat
	}
	
    /**
    Gibt den Fehlerstring in das Weltfenster aus.
    - parameters:
    	- was: Auszugebende Fehlermeldung
    */
	func fehlerMelden(was: String)
	{
		weltFenster.fehlerAusgabe(text: was)
	}
	
	/**
    zeichnet die Welt im normalen Updatezyklus
    */
	func paintWorld()
	{
		weltFenster.zeichnen();
	}
    
    /**
    Zeichent die Welt sofort
    */
    func sofortZeichnen ()
    {
        #if os(macOS)
        weltFenster.display()
        weltFenster.weltAnzeige.needsDisplay = true
        #elseif os(iOS)
        //weltFenster.display(nil)
        weltFenster.weltAnzeige.setNeedsDisplay()
        #endif
    }

	func isInside(x: Int, y: Int) -> Bool
	{
		return (x >= 1) && (x <= areaBreite) && (y >= 1) && (y <= areaLaenge)
	}

	/*
	 * legt Ziegel oder Quader an der Stelle x,y ab
	 */
	func push (x: Int, y: Int,  what: Int8, f: BildFarben)
	{
		if (what == st_Ziegel) && (areaStapelInhalt[x][y].count < areaHoehe) && (areaStapelInhalt[x][y].count == 0  || areaStapelInhalt[x][y][0].was == st_Ziegel)
		{
			areaStapelInhalt[x][y].append((st_Ziegel, f))
		}
		if (what == st_Quader) && (areaStapelInhalt[x][y].count == 0) && !(areaMarkiert[x][y].ist)
		{
			areaStapelInhalt[x][y].append((st_Quader, f))
			areaStapelInhalt[x][y].append((st_Quader, f))
		}
	}
	
	/*
	 * entfernt Ziegel oder Quader an der Stelle x,y
	 */
	func pop(x: Int, y: Int)
	{
		if areaStapelInhalt[x][y].count > 0
		{
			if (areaStapelInhalt[x][y][0].was == st_Ziegel)
			{
				areaStapelInhalt[x][y].removeLast()
			}
			else if (areaStapelInhalt[x][y][0].was == st_Quader)
			{
				areaStapelInhalt[x][y].removeLast()
				areaStapelInhalt[x][y].removeLast()
			}
		}
	}
	
	/*
	 * gibt den Gegenstand an der Stelle x,y in der Höhe z
	 */
	func getPart(x: Int, y: Int, z: Int) -> (was: Int8, farbe: BildFarben)
	{
		var ergeb: (Int8, BildFarben)
		
		if x<=0 || x>areaBreite || y<=0 || y>areaLaenge
        {
			ergeb = (st_Nichts, BildFarben.keine)
		}
        else
		{
			if z < 0 || z >= areaStapelInhalt[x][y].count
            {
				ergeb = (st_Nichts, BildFarben.keine)
            }
			else
            {
				ergeb = areaStapelInhalt[x][y][z]
            }
		}
		return ergeb
	}

	/*
	 * Anzahl der Ziegel an der Stelle x,y
	 */
	func brickCount(x: Int, y: Int) -> Int
    {
		return getPart(x: x,y: y,z: 0).was == st_Ziegel ? areaStapelInhalt[x][y].count : 0
	}

	/*
	 * prüft, ob die maximale Ziegelhöhe erreicht wurde.
	 */
	func isMaxTop(x: Int, y: Int) -> Bool
	{
		return areaStapelInhalt[x][y].count >= areaHoehe
	}

	/*
	 * Ist an dieser Stelle ein Quader?
	 */
	func isStone(x: Int, y: Int) -> Bool
    {
		return getPart(x: x,y: y,z: 0).was == st_Quader
	}

	/*
	 * Ist an dieser Stelle ein Ziegel?
	 */
	func isBrick(x: Int, y: Int) -> Bool
	{
		return getPart(x: x,y: y,z: 0).was == st_Ziegel
	}

	/*
	 * Ist an dieser Stelle ein Ziegel mit der angegebenen Farbe?
	 */
	func isBrick(x: Int, y: Int, f: BildFarben) -> Bool
	{
		return getPart(x: x,y: y,z: 0).was == st_Ziegel && getPart(x: x,y: y,z: 0).farbe == f

	}

	/*
	 * setzt Marke an der Stelle x,y
	 */
	func setMarker(x: Int, y: Int, f: BildFarben)
	{
        areaMarkiert[x][y] = (true, f)
	}

	/*
	 * entfernt Marke an der Stelle x,y
	 */
	func deleteMarker(x: Int, y: Int)
	{
        areaMarkiert[x][y] = (false, BildFarben.keine)
	}

	/*
	 * ist an der Stelle x,y eine Marke
	 */	
	func isMarker(x: Int, y: Int) -> Bool
	{
		return areaMarkiert[x][y].ist
	}

	/*
	 * ist an der Stelle x,y eine Marke mit der angegebenen Farbe
	 */	
	func isMarker(x: Int, y: Int, f: BildFarben) -> Bool
	{
		return areaMarkiert[x][y].ist && areaMarkiert[x][y].farbe == f
	}

	/*
	 * Holt die Markierungsinformation an der Stelle x,y
	 */	
	func getMarker(x: Int, y: Int) -> (ist: Bool, farbe:BildFarben)
	{
		return areaMarkiert[x][y]
	}

	/*
	 * steht an dieser Stelle ein Roboter?
	 * 0 = keiner; sonst ID des Roboters
	 */
	func getRobotID(x: Int, y: Int) -> Int
    {
		var ergeb: Int = 0
		
		for r in alleRoboter
		{
			if (r.PositionXGeben() == x) && (r.PositionYGeben() == y)
            {
				ergeb = r.KennungGeben()
            }
		}
		return ergeb
	}
	
	/*
	 * ist ein Roboter zu sehen? Der Blick geht dabei von der Stelle in die
	 * angegebene Blickrichtung. Es darf sich kein anderes Objekt zwischen
	 * der Stelle und dem Roboter befinden.
	 * 
	 */
	func isRobotInSight(xp: Int, yp: Int, blickrichtung: Character) -> Bool
	{
		var ergeb: Bool = false
		var abbruch: Bool = false
        var x = xp
        var y = yp
		
		while !ergeb && !abbruch && isInside(x: x,y: y)
		{
			switch blickrichtung
			{
				case "S" :
                	y += 1
				case "W" :
                	x -= 1
				case "N" :
                	y -= 1
				case "O" :
                	x += 1
				default :
                	y += 1
			}
			
			if isInside(x: x,y: y)
			{
				if getRobotID(x: x,y: y)>0
                {
					ergeb = true
                }
				else
                {
					abbruch = isBrick(x: x,y: y) || isStone(x: x,y: y)
                }
			}
		}
		return ergeb
	}
}

/**
Realisierung eines Roboters Karol in Java

RobotKarol ist eine Programmiersprache, die für Schülerinnen und Schüler 
zum Erlernen des Programmierens und zur Einführung in die Algorithmik gedacht ist.
Für die Entwicklung von RobotKarol-Programmen steht eine eigene Programmierumgebung
zur Verfügung (www.schule.bayern.de/karol).
JavaKarol ist eine Weiterführung dieses Konzepts bei dem die Objekte einer Karolwelt
mit Java erzeugt und bearbeitet werden können.

Die Klasse Roboter ist eine Umsetzung des Konzepts von RobotKarol in die
Programmiersprache Java. Es stehen die gleichen Methoden wie bei RobotKarol
zur Verfügung. In einer Welt können sich mehrere Objekte der Klasse Roboter bewegen
(Z.Zt. maximal 9). Auf einer Quadratkachel der Welt kann aber immer nur ein
Roboter stehen.

Methoden zum Einstellen des Roboters:
VerzoegerungSetzen(); SprungHoeheSetzen(); RucksackMaximumSetzen();
UnsichtbarMachen(); SichtbarMachen();

Methoden:
Schritt(); LinksDrehen(); RechtsDrehen(); Hinlegen(); Aufheben();
QuaderAufstellen(); QuaderEntfernen(); MarkeSetzen(); MarkeLoeschen();
TonErzeugen(); Warten(); MeldungAusgeben();

Logische Abfragefunktionen:
IstWand(); IstZiegel(); IstMarke(); IstRoboter(); IstRoboterInSicht();
IstBlickNorden(); IstBlickSueden(); IstBlickWesten(); IstBlickOsten();
IstRucksackVoll(); IstRucksackLeer(); HatZiegelImRucksack();

Attributabfragefunktionen:
PositionXGeben(); PositionYGeben(); BlickrichtungGeben(); KennungGeben();
SprunghoeheGeben();

Abfragefunktionen:
AnzahlZiegelRucksackGeben(); AnzahlZiegelVorneGeben();
RoboterVorneKennungGeben();


- author: Ulli Freiberger
- version: 1.0 (9.1.2008)

Umsetzung der Klassen nach Swift für die Entwicklung mit Xcode.
- author: Albert Wiedemann
- version: 1.0
- version: 3.0 (23.09.2019) Änderngen JavaKarol 3.0
*/
open class Roboter
{
	private struct Punkt
    {
    	var x, y: Int
        
        init (x: Int, y: Int)
        {
        	self.x = x
            self.y = y
        }
	}
    
    private var positionX: Int = 1				// Koordinaten im Raum positionX: 1 .. Breite; positionY: 1 .. Länge;
	private var positionY: Int = 1
	private var blickrichtung: Character = "S"   // Blickrichtung des Roboters
    
	private var meineWelt: Welt
	
	private var sprunghoehe: Int = 1			// maximale Sprunghöhe; default 1
	private var rucksackInhalt: Int = 0		    // Anzahl der Ziegelsteine im Rucksack; default 0
	private var rucksackMaximum: Int = 0    	// maximale Zahl der Ziegelsteine im Rucksack; >0 schaltet Überprüfung ein
	private var rucksackPruefen: Bool = false	// beim Aufheben oder Hinlegen wird der Inhalt des Rucksacks verändert und kontrolliert
	private var verzoegerung: Int = 300	        // Verzögerung nach jedem Methodenaufruf um msec; default 0,3 sec

	private var kennung: Int = 0				// eindeutige Kennung in der Karolwelt;
	private var sichtbar: Bool = true			// unsichtbare Roboter werden in der Welt nicht gezeichnet;
	
	private var startX = 1, startY: Int	= 1		// Startwerte beim Initialisieren; für reload() benötigt;
	private var startBlickrichtung: Character = "S"
    
    /**
    Initialisierung das Roboters, verankern in der Welt
    - parameters:
        - x: x-Koordinate des Startpunkts
        - y: y-Koordinate des Startpunkts
        - blickrichtung: Blickrichtung beim Start
        - inWelt: Welt, in der der Roboter wohnen soll
    */
    public init (x: Int, y: Int, blickrichtung: Character, inWelt: Welt)
    {
        meineWelt = inWelt
        StartSetzen(startX: x, startY: y, startBlickrichtung: blickrichtung)
    }
    
    /**
    Erzeugen eines neuen Roboters aufgrund der Einstellungen in einer Karolwelt-Datei
    
    Vorher muss ein Objekt der Klasse Welt angelegt werden, in der der Roboter leben soll.
    Wenn dieses Weltobjekt aus einer Karolwelt-Datei geladen wurde, dann werden
    die dortigen Robotereinstellungen verwendet, sonst ein Default-Roboter
    mit (1,1) und Blickrichtung 'S'
    - parameters:
        - inWelt: Welt, in der der Roboter wohnen soll
    */
	public init (inWelt: Welt)
	{
        meineWelt = inWelt
		let daten = meineWelt.roboterDatenAbholen()
        StartSetzen (startX: daten.posX, startY: daten.posY, startBlickrichtung: daten.direct)
	}
    
    /**
    Setzt die Startkoordinaten und meldet den Roboter in der Welt an.
    - parameters:
        - startX: x-Koordinate des Startpunkts
        - startY: y-Koordinate des Startpunkts
        - startBlickrichtung: Blickrichtung beim Start
    */
    func StartSetzen (startX: Int, startY: Int, startBlickrichtung: Character)
    {
    	positionX = max (startX, 1)
    	positionY = max (startY, 1)
        blickrichtung = (startBlickrichtung == "W") || (startBlickrichtung == "S") || (startBlickrichtung == "O") || (startBlickrichtung == "N") ? startBlickrichtung : "S"
		self.startX = startX
		self.startY = startY
		self.startBlickrichtung = startBlickrichtung
		kennung = meineWelt.roboterAnmelden(anmeldeRoboter: self,neu: true)
		meineWelt.paintWorld()
		wartenIntern()
    }

	/**
    Gibt den Zustand des Roboters als String aus.
    - returns: Namen und Position des Roboters, gegebenenfalls auch den Rucksackinhalt
    */
	open func toString() -> String
	{
		let teil1 = "Roboter \(kennung) an der Stelle (\(positionX);\(positionY)) mit Blickrichtung \(blickrichtung)"
		let teil2 = " und Rucksackinhalt \(rucksackInhalt)"
		
		if rucksackPruefen
        {
			return teil1 + teil2
        }
		else
        {
			return teil1
        }
	}
	/*
	 * Methoden zur Einstellung des Roboters
	 */

    /**
    Setzen der Verzögerung

    Nach jedem Methodenaufruf des Roboters wird um die angegebene
    Zeit gewartet.
    Defaultwert 300
    - parameters:
        - msec: Verzögerungszeit in Millisekunden
    */
	open func VerzoegerungSetzen(msec: Int)
	{
		verzoegerung = abs (msec)
	}

    /**
    Setzen der maximalen Sprunghöhe des Roboters

    Defaultwert 1
    - parameters:
        - hoehe: Sprunghöhe in Ziegeleinheiten
    */
	open func SprunghoeheSetzen(hoehe neueHoehe: Int)
	{
		sprunghoehe = max(neueHoehe, 1)
	}

    /**
    Maximales Fassungsvermögen des Rucksacks setzen

    Durch Setzen auf einen Werten größer 0 wird die
    Prüfung des Rucksackinhaltes eingeschalten. Bei jedem
    Hinlegen und Aufheben von Ziegeln wird dann
    der Rucksackinhalt geprüft.
    Defaultwert 0
    - parameters:
        - maxZiegel: Maximale Anzahl der Ziegel im Rucksack.
    */
	open func RucksackMaximumSetzen(maxZiegel: Int)
	{
		rucksackMaximum = abs (maxZiegel)
		rucksackPruefen = rucksackMaximum > 0
	}

	/**
    Unsichtbarmachen
    
    Der Roboter wird bei der Welt abgemeldet. Ab diesem
    Zeitpunkt wird sein Grafiksymbol nicht mehr in der Karolwelt
    angezeigt. Er ist nicht mehr zu sehen und hat keine Wirkung
    auf die anderen Roboter.
    */
	open func UnsichtbarMachen()
	{
		if sichtbar
		{
			meineWelt.roboterAbmelden(abmeldeRoboter: self)
			sichtbar = false
			meineWelt.paintWorld()
		}
	}
	
	/**
    Sichtbarmachen
    
    Ein vorher unsichtbar gemachter Roboter wird wieder sichtbar.
    Der Roboter muss vorher schon einmal sichtbar gewesen sein.
    */
	open func SichtbarMachen()
	{
		if !sichtbar
		{
			_ = meineWelt.roboterAnmelden(anmeldeRoboter: self,neu: false)
			sichtbar = true
			meineWelt.paintWorld()
		}
	}
	
	
	/*
	 * Anweisungen in der Sprache Karol
	 */
	
	/**
    Der Roboter macht einen Schritt nach vorne
    
	- throws: Der Schritt ist aufgrund seiner Position
            oder des Weltzustands nicht möglich
    */
    func SchrittIntern() throws
	{
		let neuX = getPositionVorne().x
		let neuY = getPositionVorne().y
		
		if !meineWelt.isInside(x: neuX, y: neuY)
		{
			meineWelt.fehlerMelden(was: toString() + " ist an der Wand angestoßen.")
            throw RoboterFehler.bewegungNichtMoeglich
		}
		if meineWelt.isStone(x: neuX, y: neuY)
		{
			meineWelt.fehlerMelden(was: toString() + " ist am Quader angestoßen.")
            throw RoboterFehler.bewegungNichtMoeglich
		}
		if meineWelt.getRobotID(x: neuX, y: neuY) > 0
		{
			meineWelt.fehlerMelden(was: toString() + " ist an anderem Roboter angestoßen.")
            throw RoboterFehler.bewegungNichtMoeglich
		}
		if abs (meineWelt.brickCount(x: positionX, y: positionY) - meineWelt.brickCount(x: neuX, y: neuY)) > sprunghoehe
		{
			meineWelt.fehlerMelden(was: toString() + " kann nicht so hoch/tief springen.")
            throw RoboterFehler.sprungHoehe
		}
		positionX = neuX
		positionY = neuY
		meineWelt.paintWorld()
		wartenIntern()
	}

	/**
    Der Roboter macht einen Schritt nach vorne
    */
	open func Schritt()
	{
    	do
        {
        	try SchrittIntern()
        }
        catch
        {
        }
        
	}
	
	/**
    Der Roboter dreht sich nach Links (gegen den Uhrzeigersinn)
    */
	open func LinksDrehen()
	{
		switch (blickrichtung)
        {
           case "N":
               blickrichtung = "W"
           case "S":
               blickrichtung = "O"
           case "O":
               blickrichtung = "N"
           case "W":
               blickrichtung = "S"
            default:
            	blickrichtung = "S"
		}
		meineWelt.paintWorld();
		wartenIntern();
	}
	
	/**
    Der Roboter dreht sich nach Rechts (im Uhrzeigersinn)
    */
	open func RechtsDrehen()
	{
		switch (blickrichtung)
        {
           case "N":
               blickrichtung = "O"
           case "S":
               blickrichtung = "W"
           case "O":
               blickrichtung = "S"
           case "W":
               blickrichtung = "N"
            default:
            	blickrichtung = "S"
		}
		meineWelt.paintWorld()
		wartenIntern()
	}

	
	/**
    Der Roboter legt vor sich einen Ziegel hin
    
    - parameters:
    	- f: Farbe des Ziegels
    - throws: Ein Hinlegen eines Ziegel ist aufgrund seiner Position
	        oder des Weltzustands nicht möglich
    - throws: Bei eingeschaltener Rucksackprüfung: wenn der Rucksack leer ist
    */
    func HinlegenIntern(f: BildFarben) throws
	{
		let vorneX = getPositionVorne().x;
		let vorneY = getPositionVorne().y;
		
		if !meineWelt.isInside(x: vorneX, y: vorneY)
		{
			meineWelt.fehlerMelden(was: toString() + " kann nicht hinlegen, er steht vor der Wand.")
			throw RoboterFehler.hinlegenNichtMoeglich
		}
		if meineWelt.isStone(x: vorneX, y: vorneY)
		{
			meineWelt.fehlerMelden(was: toString() + " kann nicht hinlegen, er steht vor einem Quader.")
			throw RoboterFehler.hinlegenNichtMoeglich
		}
		if meineWelt.getRobotID(x: vorneX, y: vorneY)>0
		{
			meineWelt.fehlerMelden(was: toString() + " kann nicht hinlegen, er steht vor einem anderen Roboter.");
			throw RoboterFehler.hinlegenNichtMoeglich
		}
		if meineWelt.isMaxTop(x: vorneX, y: vorneY)
		{
			meineWelt.fehlerMelden(was: toString() + " kann nicht hinlegen, die maximale Stapelhöhe ist erreicht.");
			throw RoboterFehler.hinlegenNichtMoeglich
		}
		if rucksackPruefen
		{
			if rucksackInhalt<1
			{
				meineWelt.fehlerMelden(was: toString() + " hat keine Ziegel mehr zum Hinlegen.")
				throw RoboterFehler.hinlegenNichtMoeglich
			}
			else
            {
				rucksackInhalt -= 1
            }
		}
		meineWelt.push(x: vorneX, y: vorneY, what: meineWelt.st_Ziegel, f: f)
		meineWelt.paintWorld()
		wartenIntern()
	}
	
	/**
    Der Roboter legt vor sich einen Ziegel hin
    Die Farbe des Ziegels ist rot
    */
	open func Hinlegen()
	{
    	do
        {
        	try HinlegenIntern(f: BildFarben.rot)
        }
        catch
        {
        }
	}
	
	/**
    Der Roboter legt vor sich einen Ziegel hin
    - parameters:
    	- farbe: Farbe des Ziegels
    */
	open func Hinlegen(farbe: String)
	{
    	do
        {
        	try HinlegenIntern(f: BildFarben.FarbeFürString(farbe: farbe))
        }
        catch
        {
        }
	}

	
	/**
    Der Roboter hebt vor sich einen Ziegel auf
    
    - throws: Ein Hinlegen eines Ziegel ist aufgrund seiner Position
	        oder des Weltzustands nicht möglich
    - throws: Bei eingeschaltener Rucksackprüfung: wenn der Rucksack leer ist
    */
    func AufhebenIntern() throws
	{
		let vorneX = getPositionVorne().x
		let vorneY = getPositionVorne().y
		
		if !meineWelt.isInside(x: vorneX, y: vorneY)
		{
			meineWelt.fehlerMelden(was: toString() + " kann nicht aufheben, er steht vor der Wand.")
			throw RoboterFehler.aufhebenNichtMoeglich
		}
		if meineWelt.isStone(x: vorneX, y: vorneY)
		{
			meineWelt.fehlerMelden(was: toString() + " kann nicht aufheben, er steht vor einem Quader.")
			throw RoboterFehler.aufhebenNichtMoeglich
		}
		if meineWelt.getRobotID(x: vorneX, y: vorneY)>0
		{
			meineWelt.fehlerMelden(was: toString() + " kann nicht aufheben, er steht vor einem anderen Roboter.")
			throw RoboterFehler.aufhebenNichtMoeglich
		}
		if meineWelt.brickCount(x: vorneX, y: vorneY)<1
		{
			meineWelt.fehlerMelden(was: toString() + " kann nicht aufheben, vor ihm liegen keine Ziegel.")
			throw RoboterFehler.aufhebenNichtMoeglich
		}
		if rucksackPruefen
		{
			if rucksackInhalt >= rucksackMaximum
			{
				meineWelt.fehlerMelden(was: toString() + " kann nicht aufheben, das maximale Tragvermögen ist erreicht.")
				throw RoboterFehler.aufhebenNichtMoeglich
			}
			else
            {
				rucksackInhalt += 1
            }
		}
		meineWelt.pop(x: vorneX, y: vorneY)
		meineWelt.paintWorld()
		wartenIntern()
	}
	
	/**
    Der Roboter hebt vor sich einen Ziegel auf.
    */
	open func Aufheben()
	{
    	do
        {
            try AufhebenIntern()
        }
        catch
        {
        	//assertionFailure("Es kann kein Ziegel aufgehoben werden.")
        }
	}

	
	/**
	 Der Roboter stellt vor sich einen Quader auf
	 
	 - throws: Das Aufstellen eines Quaders ist aufgrund seiner Position
            oder des Weltzustands nicht möglich
    */
    func QuaderAufstellenIntern() throws
	{
		let vorneX = getPositionVorne().x
		let vorneY = getPositionVorne().y
		
		if !meineWelt.isInside(x: vorneX, y: vorneY)
		{
			meineWelt.fehlerMelden(was: toString() + " kann Quader nicht aufstellen, er steht vor der Wand.")
			throw RoboterFehler.aufstellenNichtMoeglich
		}
		if meineWelt.isStone(x: vorneX, y: vorneY)
		{
			meineWelt.fehlerMelden(was: toString() + " kann Quader nicht aufstellen, er steht vor einem Quader.")
			throw RoboterFehler.aufstellenNichtMoeglich
		}
		if meineWelt.getRobotID(x: vorneX, y: vorneY)>0
		{
			meineWelt.fehlerMelden(was: toString() + " kann Quader nicht aufstellen, er steht vor einem anderen Roboter.")
			throw RoboterFehler.aufstellenNichtMoeglich
		}
		if meineWelt.brickCount(x: vorneX, y: vorneY)>0
		{
			meineWelt.fehlerMelden(was: toString() + " kann Quader nicht aufstellen, vor ihm liegen Ziegel.")
			throw RoboterFehler.aufstellenNichtMoeglich
		}
		if meineWelt.isMarker(x: vorneX, y: vorneY)
		{
			meineWelt.fehlerMelden(was: toString() + " kann Quader nicht aufstellen, vor ihm ist eine Marke.")
			throw RoboterFehler.aufstellenNichtMoeglich
		}
		meineWelt.push(x: vorneX,y: vorneY,what: meineWelt.st_Quader, f: BildFarben.keine)
		meineWelt.paintWorld()
		wartenIntern()
	}
	
	/**
    Der Roboter stellt vor sich einen Quader auf
    */
	open func QuaderAufstellen()
	{
    	do
        {
            try QuaderAufstellenIntern()
        }
        catch
        {
        }
	}

	/**
    Der Roboter entfernt einen Quader der vor ihm steht
    
    - throws: Das Entfernen des Quaders ist aufgrund seiner Position
            oder des Weltzustands nicht möglich
    */
    func QuaderEntfernenIntern() throws
	{
		let vorneX = getPositionVorne().x
		let vorneY = getPositionVorne().y
		
		if !meineWelt.isInside(x: vorneX, y: vorneY)
		{
			meineWelt.fehlerMelden(was: toString() + " kann Quader nicht entfernen, er steht vor der Wand.")
			throw RoboterFehler.abbauenNichtMoeglich
		}
		if meineWelt.getRobotID(x: vorneX, y: vorneY)>0
		{
			meineWelt.fehlerMelden(was: toString() + " kann Quader nicht entfernen., er steht vor einem anderen Roboter.")
			throw RoboterFehler.abbauenNichtMoeglich
		}
		if !meineWelt.isStone(x: vorneX, y: vorneY)
		{
			meineWelt.fehlerMelden(was: toString() + " kann Quader nicht entfernen., er steht vor keinem Quader.")
			throw RoboterFehler.abbauenNichtMoeglich
		}
		meineWelt.pop(x: vorneX, y: vorneY)
		meineWelt.paintWorld()
		wartenIntern()
	}
	
	/**
    Der Roboter entfernt einen Quader der vor ihm steht
    */
	open func QuaderEntfernen()
	{
    	do
        {
            try QuaderEntfernenIntern()
        }
        catch
        {
        }
	}

	/**
    Der Roboter setzt auf der Kachel unter ihm eine Marke
    */
	open func MarkeSetzen()
	{
		meineWelt.setMarker(x: positionX, y: positionY, f: BildFarben.gelb)
		meineWelt.paintWorld()
		wartenIntern()
	}

	/**
    Der Roboter setzt auf der Kachel unter ihm eine Marke
    - parameters:
    	- farbe: die Farbe der Marke
    */
	open func MarkeSetzen(farbe: String)
	{
		meineWelt.setMarker(x: positionX, y: positionY, f: BildFarben.FarbeFürString(farbe: farbe))
		meineWelt.paintWorld()
		wartenIntern()
	}

	/**
    Der Roboter entfernt von der Kachel unter ihm eine Marke
    */
	open func MarkeLoeschen()
	{
		meineWelt.deleteMarker(x: positionX, y: positionY)
		meineWelt.paintWorld()
		wartenIntern()
	}
	
	/**
    Der Roboter gibt einen Piep-Ton aus
    */
	open func TonErzeugen()
	{
        //AudioServicesPlayAlertSound (kSystemSoundID_Vibrate)
	}

    /**
    Der Roboter wartet die angegebene Zeit
    - parameters:
        - dauer: Wartezeit in Sekunden
    */
	open func Warten(dauer: Float)
	{
		Thread.sleep(forTimeInterval: TimeInterval(dauer))
	}

	/**
	 Der Roboter gibt eine Meldung an das Weltfenster aus
	 - parameters:
	 	- text: Der auszugebende Meldetext
    */
	open func MeldungAusgeben(text was: String)
	{
		meineWelt.fehlerMelden(was: toString() + " sagt: " + was)
	}


	/*
	 * Bedingungen in der Sprache Karol
	 * Abfragen, die mit true oder false beantwortet werden
	 */
	
	/**
    Abfrage ob der Roboter vor einer Wand bzw. einem Quader steht
	 
    - returns: wenn zutrifft dann true
    */
	open func IstWand() -> Bool
	{
		let vorneX = getPositionVorne().x
		let vorneY = getPositionVorne().y
		return !meineWelt.isInside(x: vorneX, y: vorneY) || (meineWelt.isStone(x: vorneX, y: vorneY))
	}

	/**
    Abfrage ob vor dem Roboter ein Ziegel liegt
	 
    - returns: wenn zutrifft dann true
    */
	open func IstZiegel() -> Bool
	{
		let vorneX = getPositionVorne().x
		let vorneY = getPositionVorne().y
		return meineWelt.isBrick(x: vorneX, y: vorneY)
	}

	/**
    Abfrage ob vor dem Roboter ein Ziegel der angegbenen Farbe liegt
    - parameters:
    	- farbe: Farbe, die der Ziegel haben soll
    - returns: wenn zutrifft dann true
    */
	open func IstZiegel(farbe f: String) -> Bool
	{
		let vorneX = getPositionVorne().x
		let vorneY = getPositionVorne().y
		return meineWelt.isBrick(x: vorneX, y: vorneY, f: BildFarben.FarbeFürString(farbe: f))
	}

    /**
    Abfrage ob vor dem Roboter ein Ziegel liegt
     
    - returns: wenn zutrifft dann true
    */
    open func IstZiegelLinks() -> Bool
    {
        let vorneX = getPositionLinks().x
        let vorneY = getPositionLinks().y
        return meineWelt.isBrick(x: vorneX, y: vorneY)
    }

    /**
    Abfrage ob vor dem Roboter ein Ziegel liegt
     
    - returns: wenn zutrifft dann true
    */
    open func IstZiegelRechts() -> Bool
    {
        let vorneX = getPositionRechts().x
        let vorneY = getPositionRechts().y
        return meineWelt.isBrick(x: vorneX, y: vorneY)
    }

	/**
    Abfrage ob sich unter dem Roboter eine Marke befindet
    
    - returns: wenn zutrifft dann true
    */
	open func IstMarke() -> Bool
	{
		return meineWelt.isMarker(x: positionX, y: positionY)
	}
	
	/**
    Abfrage ob sich unter dem Roboter eine Marke der angegbenen Farbe befindet
    - parameters:
    	- farbe: Farbe, die die Marke haben soll
    - returns: wenn zutrifft dann true
    */
	open func IstMarke(farbe f: String) -> Bool
	{
		return meineWelt.isMarker(x: positionX, y: positionY, f: BildFarben.FarbeFürString(farbe: f))
	}
	
	/**
    Abfrage ob direkt vor dem Roboter ein anderer Roboter steht
    
    - returns: wenn zutrifft dann true
    */
	open func IstRoboter() -> Bool
	{
		let vorneX = getPositionVorne().x
		let vorneY = getPositionVorne().y
		return meineWelt.getRobotID(x: vorneX, y: vorneY)>0
	}

	/**
    Abfrage ob sich in Blickrichtung ein anderer Roboter befindet,
    egal wie weit weg. Ziegel und Quader behindern die Sicht.
    - returns: wenn zutrifft dann true
    */
	open func IstRoboterInSicht() -> Bool
	{
		return meineWelt.isRobotInSight(xp: positionX, yp: positionY, blickrichtung: blickrichtung)
	}
	
	/**
    Abfrage ob die Blickrichtung Norden ist
    
    - returns: wenn zutrifft dann true
    */
	open func IstBlickNorden() -> Bool
	{
		return blickrichtung == "N"
	}

	/**
    Abfrage ob die Blickrichtung Süden ist
    
    - returns: wenn zutrifft dann true
    */
	open func IstBlickSueden() -> Bool
	{
		return blickrichtung == "S"
	}

	/**
    Abfrage ob die Blickrichtung Osten ist
    
    - returns: wenn zutrifft dann true
    */
	func IstBlickOsten() -> Bool
	{
		return blickrichtung == "O"
	}

	/**
    Abfrage ob die Blickrichtung Westen ist
    
    - returns: wenn zutrifft dann true
    */
	open func IstBlickWesten() -> Bool
	{
		return blickrichtung == "W"
	}
	
	/**
    Abfrage ob der Rucksack voll ist
    
    - returns: wenn zutrifft dann true
    */
	open func IstRucksackVoll() -> Bool
	{
		return rucksackInhalt >= rucksackMaximum
	}

	/**
    Abfrage ob der Rucksack leer ist
    
    - returns: wenn zutrifft dann true
    */
	open func IstRucksackLeer() -> Bool
	{
		return rucksackInhalt == 0
	}

	/**
    Abfrage ob Ziegel im Rucksack sind
    
    - returns: wenn zutrifft dann true
    */
	open func HatZiegelImRucksack() -> Bool
	{
		return rucksackInhalt > 0
	}
	
	
	/*
	 * Abfragen in der Sprache Karol
	 * und Getter-Methoden
	 */
	
	/**
    Gibt die X-Koordinate der Position
    
	- returns:  X-Wert 1..Weltbreite
    */
    open func PositionXGeben() -> Int
	{
		return positionX
	}
	
	/**
    Gibt die Y-Koordinate der Position
    
	- returns: Y-Wert 1..Weltlänge
    */
	open func PositionYGeben() -> Int
	{
		return positionY
	}
	
	/**
    Gibt die Blickrichtung
    
	- returns: Blickrichtung "S","W","N","O"
    */
	open func BlickrichtungGeben() -> Character
	{
		return blickrichtung
	}
	
	/**
    Gibt die Blickrichtung als Zahl
    
	- returns: Int "S"=0,"W"=1,"N"=2,"O"=3
    */
	func getBlickrichtungNr() -> Int
	{
		switch blickrichtung
        {
        	case "S":
            	return 0
        	case "W":
            	return 1
        	case "N":
            	return 2
        	case "O":
            	return 3
        	default:
            	assertionFailure("Unbekannte Blickrichtung \(blickrichtung) in getBlickrichtungNr")
                return 0
        }
	}

	/**
    Gibt die maximale Sprunghöhe zurück
    
	- returns: Sprunghöhe
	 */
	open func SprungshoeheGeben() -> Int
	{
		return sprunghoehe
	}

    /**
    Gibt die Rucksackprüfung zurück
    
    - returns: true = Rucksack prüfen
     */
    open func RucksackPruefungGeben() -> Bool
    {
        return rucksackPruefen
    }

	/**
    Gibt die Kennung des Roboters
    
	- returns: Kennung als Ganzzahl
    */
	open func KennungGeben() -> Int
	{
		return kennung
	}

	/**
    Gibt die Sichtbarkeit des Roboters
    
	- returns: true = sichtbar
    */
	open func SichtbarkeitGeben() -> Bool
	{
		return sichtbar
	}

	/**
    Gibt die Anzahl der Ziegel die sich im Rucksack befinden
    
	- returns: Anzahl Ziegel
    */
	open func AnzahlZiegelRucksackGeben() -> Int
	{
		return rucksackInhalt
	}

	/**
    Gibt die Anzahl der Ziegel die vor dem Roboter stehen
    
	- returns:  Anzahl Ziegel
    */
	open func AnzahlZiegelVorneGeben() -> Int
	{
		let vorneX = getPositionVorne().x
		let vorneY = getPositionVorne().y
		return meineWelt.brickCount(x: vorneX, y: vorneY)
	}
	
	
	/**
    Gibt die Kennung des Roboters der vor dem Roboter steht
    
	- returns: Kennung des Roboters,
        ist vorne kein Roboter wird 0 zurückgegeben
    */
	open func RoboterVorneKennungGeben() -> Int
	{
		let vorneX = getPositionVorne().x
		let vorneY = getPositionVorne().y
		return meineWelt.getRobotID(x: vorneX, y: vorneY)
	}


	/*
	 * Interne private Methoden
	 */
	func wartenIntern()
	{
    	let v = Double (verzoegerung) / 1000.0
        if v > 0
        {
        	meineWelt.sofortZeichnen()
        }
		Thread.sleep(forTimeInterval: TimeInterval(v))
	}
	
	private func getPositionVorne() -> Punkt
	{
		var pos: Punkt = Punkt(x: positionX, y: positionY)

		switch blickrichtung
		{
    		case "N": pos.y -= 1
    		case "S": pos.y += 1
    		case "O": pos.x += 1
    		case "W": pos.x -= 1
            default: pos.x += 0
		}
		return pos
	}
    
    private func getPositionLinks() -> Punkt
    {
        var pos: Punkt = Punkt(x: positionX, y: positionY)

        switch blickrichtung
        {
            case "N": pos.x -= 1
            case "S": pos.x += 1
            case "O": pos.y -= 1
            case "W": pos.y += 1
            default: pos.x += 0
        }
        return pos
    }
    
    private func getPositionRechts() -> Punkt
    {
        var pos: Punkt = Punkt(x: positionX, y: positionY)

        switch blickrichtung
        {
            case "N": pos.x += 1
            case "S": pos.x -= 1
            case "O": pos.y += 1
            case "W": pos.y -= 1
            default: pos.x += 0
        }
        return pos
    }

	/**
    Setzt auf den Initialisierungszustand zurück
	prüft ob der Roboter Platz hat
    wird nur von der Welt aufgerufen
    */
	func reset()
	{
		if (meineWelt.isInside(x: startX, y: startY) && 
		   (!meineWelt.isStone(x: startX, y: startY)) && 
		   (meineWelt.getRobotID(x: startX, y: startY)==0))
		{
			positionX = startX
			positionY = startY
			blickrichtung = startBlickrichtung
		}
		else
		{
			for x in 1 ..< meineWelt.areaBreite
            {
				for y in 1 ..< meineWelt.areaLaenge
                {
					if meineWelt.isInside(x: x, y: y) &&
					   (!meineWelt.isStone(x: x, y: y)) &&
					   (meineWelt.getRobotID(x: x, y: y)==0)
					{
						positionX = x
						positionY = y
						blickrichtung = startBlickrichtung
					}
                }
            }
		}
	}
}



/**
Darstellung einer Karolwelt auf einem JPanel

Wird von einem Objekt der Klasse WeltFenster direkt erzeugt.
Die Karolwelt und die darin enthaltenen Images für Ziegel,
Quader, Marken und Roboter wird in einem BufferedImage
zeichenFlaeche gezeichnet.

Die Images für Ziegel, Quader, Marke und Roboter (9 verschiedene,
jeweils 4 Blickrichtungen) befinden sich in einem Verzeichnis imgs
das im selben Verzeichnis wie die Applikation liegen muss.

- author: Ulli Freiberger
- version: 1.0 (4.1.2008)

Umsetzung der Klassen nach Swift für die Entwicklung mit Xcode.
- author: Albert Wiedemann
- version: 1.0
*/
class WeltAnzeige: View
{
	private class Punkt2D
    {
    	var x: CGFloat
        var y: CGFloat
        
        init (x: CGFloat, y: CGFloat)
        {
        	self.x = x
            self.y = y
        }
    }

	private var welt: Welt! = nil
	private var ziegel_rot: CGImage! = nil
	private var ziegel_gelb: CGImage! = nil
	private var ziegel_grün: CGImage! = nil
	private var ziegel_blau: CGImage! = nil
	private var marke_rot: CGImage! = nil
	private var marke_gelb: CGImage! = nil
	private var marke_grün: CGImage! = nil
    private var marke_blau: CGImage! = nil
    private var marke_schwarz: CGImage! = nil
	private var quader: CGImage! = nil
    private var roboter: [[CGImage]]! = nil
    private var weltOrigin3D = Punkt2D (x: 0, y: 0)
    private var punktMax = Punkt2D (x: 0, y: 0)
	private let randLinks = 40  // Rand zwischen Zeichenfläche und Weltfläche
	private let randOben  = 60
	private let randUnten = 30
	private let weltOben  = 40  // Distanz zwischen Rückwandoberkante und Weltrandoben
    private let blau = SystemFarbe(red: 0, green: 0, blue: 1, alpha: 1).cgColor
    private var asc: CGFloat = 0.0
    private var line: CTLine! = nil
    private var font: CTFont! = nil
    private var ziegelBreite: CGFloat = 0
    private var ziegelHöhe: CGFloat = 0
    private var quaderBreite: CGFloat = 0
    private var quaderHöhe: CGFloat = 0
    private var markenBreite: CGFloat = 0
    private var markenHöhe: CGFloat = 0
    
	/*
     *	Standardkonstruktor
     */
	override init (frame: CGRect)
    {
    	super.init(frame: frame)
        BilderLaden()
        Vorbesetzungen()
    }

	/*
     *	Konstruktor für Erstellung aus NIB-File
     */
	required init? (coder: NSCoder)
    {
    	super.init (coder: coder)
        BilderLaden()
        Vorbesetzungen()
    }

    private func BilderLaden ()
    {
        var url: CFURL!
        let mainBundle = CFBundleGetMainBundle ()
        url = CFBundleCopyResourceURL(mainBundle, "Ziegel_rot" as CFString, "png" as CFString, "" as CFString)
        ziegel_rot = CGImage (pngDataProviderSource: CGDataProvider (url: url)!, decode: nil, shouldInterpolate: false, intent: CGColorRenderingIntent.defaultIntent)
        url = CFBundleCopyResourceURL(mainBundle, "Ziegel_gelb" as CFString, "png" as CFString, "" as CFString)
        ziegel_gelb = CGImage (pngDataProviderSource: CGDataProvider (url: url)!, decode: nil, shouldInterpolate: false, intent: CGColorRenderingIntent.defaultIntent)
        url = CFBundleCopyResourceURL(mainBundle, "Ziegel_grün" as CFString, "png" as CFString, "" as CFString)
        ziegel_grün = CGImage (pngDataProviderSource: CGDataProvider (url: url)!, decode: nil, shouldInterpolate: false, intent: CGColorRenderingIntent.defaultIntent)
        url = CFBundleCopyResourceURL(mainBundle, "Ziegel_blau" as CFString, "png" as CFString, "" as CFString)
        ziegel_blau = CGImage (pngDataProviderSource: CGDataProvider (url: url)!, decode: nil, shouldInterpolate: false, intent: CGColorRenderingIntent.defaultIntent)
        url = CFBundleCopyResourceURL(mainBundle, "Marke_rot" as CFString, "png" as CFString, "" as CFString)
        marke_rot = CGImage (pngDataProviderSource: CGDataProvider (url: url)!, decode: nil, shouldInterpolate: false, intent: CGColorRenderingIntent.defaultIntent)
        url = CFBundleCopyResourceURL(mainBundle, "Marke_gelb" as CFString, "png" as CFString, "" as CFString)
        marke_gelb = CGImage (pngDataProviderSource: CGDataProvider (url: url)!, decode: nil, shouldInterpolate: false, intent: CGColorRenderingIntent.defaultIntent)
        url = CFBundleCopyResourceURL(mainBundle, "Marke_grün" as CFString, "png" as CFString, "" as CFString)
        marke_grün = CGImage (pngDataProviderSource: CGDataProvider (url: url)!, decode: nil, shouldInterpolate: false, intent: CGColorRenderingIntent.defaultIntent)
        url = CFBundleCopyResourceURL(mainBundle, "Marke_blau" as CFString, "png" as CFString, "" as CFString)
        marke_blau = CGImage (pngDataProviderSource: CGDataProvider (url: url)!, decode: nil, shouldInterpolate: false, intent: CGColorRenderingIntent.defaultIntent)
        url = CFBundleCopyResourceURL(mainBundle, "Marke_schwarz" as CFString, "png" as CFString, "" as CFString)
        marke_schwarz = CGImage (pngDataProviderSource: CGDataProvider (url: url)!, decode: nil, shouldInterpolate: false, intent: CGColorRenderingIntent.defaultIntent)
        url = CFBundleCopyResourceURL(mainBundle, "Quader" as CFString, "png" as CFString, "" as CFString)
        quader = CGImage (pngDataProviderSource: CGDataProvider (url: url)!, decode: nil, shouldInterpolate: false, intent: CGColorRenderingIntent.defaultIntent)
        roboter = [[CGImage]] ()
        for (j, str) in ["S", "W", "N", "O"].enumerated()
        {
        	roboter.append([CGImage]())
            for i in 1 ... Welt.maxRoboter
            {
                url = CFBundleCopyResourceURL(mainBundle, "robot\(str)\(i)" as CFString, "png" as CFString, "" as CFString)
                let img = CGImage (pngDataProviderSource: CGDataProvider (url: url)!, decode: nil, shouldInterpolate: false, intent: CGColorRenderingIntent.defaultIntent)
            	roboter [j].append(img!)
            }
        }
    }
    
    private func Vorbesetzungen ()
    {
        font = CTFontCreateWithName ("Helvetica" as CFString, 14.0, nil)
        let attrString = NSAttributedString(string: "N", attributes: [NSAttributedString.Key.foregroundColor:blau, NSAttributedString.Key.font:font!])
        line = CTLineCreateWithAttributedString(attrString)
        var desc, lead: CGFloat
        desc = 0
        lead = 0
        let _ = CTLineGetTypographicBounds (line, &asc, &desc, &lead)
    	ziegelBreite = CGFloat (ziegel_rot.width)
    	ziegelHöhe = CGFloat (ziegel_rot.height)
    	quaderBreite = CGFloat (quader.width)
    	quaderHöhe = CGFloat (quader.height)
        markenBreite = CGFloat (marke_gelb.width)
        markenHöhe = CGFloat (marke_gelb.height)
    }
    
	/*
	 * Abbildung von 3D auf 2D
	 * bezüglich den Koordinaten der Weltfläche
	 */
	private func p3(x: CGFloat, y: CGFloat, z: CGFloat) -> Punkt2D
	{
        return Punkt2D (x: weltOrigin3D.x + 30*x - 15*y, y: weltOrigin3D.y + 15*y - 15*z)
	}

	/*
	 * Bestimmt erforderliche Größe der Zeichenfläche
	 * und die endgültige Position von WeltOrigin3D
	 * aufgrund der Größe der Welt
	 */
	private func WeltGroesseErmitteln()
	{
		var links: CGFloat = CGFloat.infinity
		var oben: CGFloat = CGFloat.infinity
		var rechts: CGFloat = -CGFloat.infinity
		var unten: CGFloat = -CGFloat.infinity
		
        let h = welt.areaHoehe
        let b = welt.areaBreite
        let l = welt.areaLaenge
        let punkte = [p3(x: 0, y: 0, z: 0), p3(x: 0, y: 0, z: CGFloat(h)),
        	p3(x: 0, y: CGFloat(l), z: 0),
            p3(x: 0, y: CGFloat(l), z: CGFloat(h)),
            p3(x: CGFloat(b), y: 0, z: 0),
            p3(x: CGFloat(b), y: 0, z: CGFloat(h)),
            p3(x: CGFloat(b), y: CGFloat(l), z: 0),
            p3(x: CGFloat(b), y: CGFloat(l), z: CGFloat(h))]
		for p in punkte
		{
			if p.x > rechts
            {
				rechts = p.x
            }
			if p.y > unten
            {
				unten = p.y
            }
			if p.x < links
            {
				links = p.x
            }
			if p.y < oben
            {
				oben = p.y
            }
		}
		
        punktMax = Punkt2D (x: abs(rechts - links) + CGFloat (2*randLinks), y: abs(unten - oben) + CGFloat (randOben + randUnten + weltOben))
        weltOrigin3D = Punkt2D (x: abs(punkte[0].x - CGFloat (links)), y: abs(punkte[0].y - CGFloat (oben)) + CGFloat (weltOben))
	}

    override func draw(_ rect: CGRect) {
    	var p1, p2: Punkt2D
        #if os(macOS)
        let context = NSGraphicsContext.current!.cgContext
        #elseif os(iOS)
        let context = UIGraphicsGetCurrentContext()!
        #endif
        if welt != nil
        {
            #if os(iOS)
            context.scaleBy(x:1, y:-1)
            context.translateBy(x:0, y:-bounds.height)
            #endif
			context.clear(bounds)
            context.setFillColor(SystemFarbe(red: 1, green: 1, blue: 1, alpha: 1).cgColor)
            context.fill(bounds)
            context.stroke(CGRect(x: 0, y: 0, width: bounds.size.width - 1, height: bounds.size.height - 1))
            if welt.ist2d
            {
                context.setStrokeColor(SystemFarbe(red: 0, green: 1, blue: 0, alpha: 1).cgColor)
                context.stroke(CGRect(x: 1, y: 1, width: bounds.size.width - 3, height: bounds.size.height - 3))
                context.setStrokeColor(SystemFarbe(red: 0, green: 0, blue: 0, alpha: 1).cgColor)
                let sc = min (bounds.size.width / CGFloat ((welt.areaBreite + 1) * 40 + 20), bounds.size.height / CGFloat ((welt.areaLaenge + 1) * 40 + 20))
                context.scaleBy (x: sc, y: sc)
                let yoffs = bounds.size.height / sc - 60
                let xoffs = 20
                for a in 1 ... welt.areaBreite
                {
                    for b in 1 ... welt.areaLaenge
                    {
                    	let h = welt.areaStapelInhalt[a][b].count
                        let part = welt.getPart(x: a, y: b, z: h - 1)
                        var invers = false
                        if part.was == welt.st_Ziegel
                        {
                            switch part.farbe
                            {
                                case BildFarben.rot:
                                    context.setFillColor(SystemFarbe (red: 1, green: 0, blue: 0, alpha: 1).cgColor)
                                case BildFarben.gelb:
                                    context.setFillColor(SystemFarbe (red: 1, green: 1, blue: 0, alpha: 1).cgColor)
                                    invers = true
                                case BildFarben.grün:
                                    context.setFillColor(SystemFarbe (red: 0, green: 1, blue: 0, alpha: 1).cgColor)
                                case BildFarben.blau:
                                    context.setFillColor(SystemFarbe (red: 0, green: 0, blue: 1, alpha: 1).cgColor)
                                default:
                                    context.setFillColor(SystemFarbe (red: 1, green: 0.5, blue: 0.5, alpha: 1).cgColor)
                            }
                        	context.fill(CGRect(x: CGFloat ((a-1)*40 + xoffs), y: yoffs - CGFloat((b-1)*40), width: 40, height: 40))
                        }
                        if part.was == welt.st_Quader
                        {
                            context.setFillColor(SystemFarbe (red: 0.7, green: 0.7, blue: 0.7, alpha: 1).cgColor)
                        	context.fill(CGRect(x: CGFloat ((a-1)*40 + xoffs), y: yoffs - CGFloat((b-1)*40), width: 40, height: 40))
                        }
                        let m = welt.getMarker(x: a, y: b)
                        if m.ist
                        {
                            switch m.farbe
                            {
                                case BildFarben.rot:
                                    context.setFillColor(SystemFarbe (red: 1, green: 0, blue: 0, alpha: 1).cgColor)
                                case BildFarben.gelb:
                                    context.setFillColor(SystemFarbe (red: 1, green: 1, blue: 0, alpha: 1).cgColor)
                                    invers = true
                                case BildFarben.grün:
                                    context.setFillColor(SystemFarbe (red: 0, green: 1, blue: 0, alpha: 1).cgColor)
                                case BildFarben.blau:
                                    context.setFillColor(SystemFarbe (red: 0, green: 0, blue: 1, alpha: 1).cgColor)
                                case BildFarben.schwarz:
                                    context.setFillColor(SystemFarbe (red: 0.23, green: 0.23, blue: 0.23, alpha: 1).cgColor)
                                default:
                                    context.setFillColor(SystemFarbe (red: 1, green: 0.5, blue: 0.5, alpha: 1).cgColor)
                            }
                        	context.fill(CGRect(x: CGFloat ((a-1)*40 + xoffs + 2), y: yoffs - CGFloat((b-1)*40 - 2), width: 36, height: 36))
                        	context.stroke(CGRect(x: CGFloat ((a-1)*40 + xoffs + 2), y: yoffs - CGFloat((b-1)*40 - 2), width: 36, height: 36))
                        }
                        if h >= 1 && part.was == welt.st_Ziegel
                        {
                            var color: CGColor
                            if invers
                            {
                            	color = SystemFarbe(red: 0, green: 0, blue: 0, alpha: 1).cgColor
                            }
                            else
                            {
                            	color = SystemFarbe(red: 1, green: 1, blue: 1, alpha: 1).cgColor
                            }
                            let attrString = NSAttributedString(string: "\(h)", attributes: [NSAttributedString.Key.foregroundColor:color, NSAttributedString.Key.font:font!])
                            let line = CTLineCreateWithAttributedString(attrString)
                            context.textPosition = CGPoint(x: CGFloat ((a-1)*40 + 16 + xoffs), y: yoffs - CGFloat((b-1)*40 - 15))
                            CTLineDraw (line, context)
                        }

                        for robo in welt.alleRoboter
                        {
                            if (a == robo.PositionXGeben()) && (b == robo.PositionYGeben()) && robo.SichtbarkeitGeben()
                            {
                            	context.beginPath()
                                switch robo.BlickrichtungGeben()
                                {
                                	case "N":
                                    	context.move(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 4), y: yoffs - CGFloat((b-1)*40 - 4)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 36), y: yoffs - CGFloat((b-1)*40 - 4)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 20), y: yoffs - CGFloat((b-1)*40 - 36)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 4), y: yoffs - CGFloat((b-1)*40 - 4)))
                                	case "S":
                                    	context.move(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 4), y: yoffs - CGFloat((b-1)*40 - 36)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 36), y: yoffs - CGFloat((b-1)*40 - 36)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 20), y: yoffs - CGFloat((b-1)*40 - 4)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 4), y: yoffs - CGFloat((b-1)*40 - 36)))
                                	case "O":
                                    	context.move(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 4), y: yoffs - CGFloat((b-1)*40 - 4)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 36), y: yoffs - CGFloat((b-1)*40 - 20)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 4), y: yoffs - CGFloat((b-1)*40 - 36)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 4), y: yoffs - CGFloat((b-1)*40 - 4)))
                                	case "W":
                                    	context.move(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 36), y: yoffs - CGFloat((b-1)*40 - 4)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 4), y: yoffs - CGFloat((b-1)*40 - 20)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 36), y: yoffs - CGFloat((b-1)*40 - 36)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 36), y: yoffs - CGFloat((b-1)*40 - 4)))
                                    default:
                                     	context.move(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 4), y: yoffs - CGFloat((b-1)*40 - 4)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 36), y: yoffs - CGFloat((b-1)*40 - 4)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 36), y: yoffs - CGFloat((b-1)*40 - 36)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 4), y: yoffs - CGFloat((b-1)*40 - 36)))
                                    	context.addLine(to: CGPoint(x: CGFloat ((a-1)*40 + xoffs + 4), y: yoffs - CGFloat((b-1)*40 - 2)))
                                }
                                context.setFillColor(SystemFarbe (red: 0.1, green: 0.1, blue: 0.1, alpha: 1).cgColor)
                                context.fillPath()
                                if robo.KennungGeben() != 1
                                {
                                    let attrString = NSAttributedString(string: "\(robo.KennungGeben())", attributes: [NSAttributedString.Key.foregroundColor:SystemFarbe(red: 1, green: 1, blue: 1, alpha: 1).cgColor, NSAttributedString.Key.font:font!])
                                    let line = CTLineCreateWithAttributedString(attrString)
                                    context.textPosition = CGPoint(x: CGFloat ((a-1)*40 + 16 + xoffs), y: yoffs - CGFloat((b-1)*40 - 15))
                                    CTLineDraw (line, context)
                                }
                            }
                            
                            context.setStrokeColor (blau)
                            context.beginPath()
                            for a in 0 ... welt.areaBreite
                            {
                            	context.move(to: CGPoint(x: CGFloat (a*40 + xoffs), y: yoffs + 40))
                                context.addLine(to: CGPoint(x: CGFloat (a*40 + xoffs), y: yoffs - CGFloat((welt.areaLaenge - 1)*40)))
                            }
                            for b in 0 ... welt.areaLaenge
                            {
                            	context.move(to: CGPoint(x: CGFloat (xoffs), y: yoffs - CGFloat((b-1)*40)))
                                context.addLine(to: CGPoint(x: CGFloat (welt.areaBreite*40 + xoffs), y: yoffs - CGFloat((b-1)*40)))
                            }
                            context.strokePath()
                        }
                    }
                }
            }
            else
            {
                let dashes: [CGFloat] = [10, 5, 5, 5]
                let sc = min (bounds.size.width / punktMax.x, bounds.size.height / punktMax.y)
                let yoffs = bounds.size.height / sc
                let xoffs = 10 / sc
                context.setStrokeColor(SystemFarbe(red: 1, green: 1, blue: 0, alpha: 1).cgColor)
                context.stroke(CGRect(x: 1, y: 1, width: bounds.size.width - 3, height: bounds.size.height - 3))
                context.scaleBy (x: sc, y: sc)
                context.textMatrix  = CGAffineTransform(a: sc, b: 0.0, c: 0.0, d: sc, tx: 0.0, ty: 0.0)
                context.setStrokeColor (blau)
                
                // Boden
                context.setLineDash(phase: 0, lengths: [])
                for i in 0 ... welt.areaLaenge
                {
                    let p1 = p3(x: 0, y: CGFloat (i), z: 0)
                    let p2 = p3(x: CGFloat(welt.areaBreite), y: CGFloat(i), z: 0)
                    context.move(to: CGPoint(x: xoffs + p1.x, y: yoffs - p1.y))
                    context.addLine(to: CGPoint(x: xoffs + p2.x, y: yoffs - p2.y))
                    
                }
                for i in 0 ... welt.areaBreite
                {
                    let p1 = p3(x: CGFloat (i), y: 0, z: 0)
                    let p2 = p3(x: CGFloat(i), y: CGFloat(welt.areaLaenge), z: 0)
                    context.move(to: CGPoint(x: xoffs + p1.x, y: yoffs - p1.y))
                    context.addLine(to: CGPoint(x: xoffs + p2.x, y: yoffs - p2.y))
                    
                }
                context.strokePath()
            
                // Rückwand
                context.setLineDash(phase: 0, lengths: dashes)
                for i in 0 ... welt.areaBreite
                {
                    let p1 = p3(x: CGFloat (i), y: 0, z: 0)
                    let p2 = p3(x: CGFloat(i), y: 0, z: CGFloat(welt.areaHoehe))
                    context.move(to: CGPoint(x: xoffs + p1.x, y: yoffs - p1.y))
                    context.addLine(to: CGPoint(x: xoffs + p2.x, y: yoffs - p2.y))
                    
                }
                p1 = p3(x: CGFloat (welt.areaBreite), y: 0, z: CGFloat(welt.areaHoehe))
                p2 = p3(x: 0, y: 0, z: CGFloat(welt.areaHoehe))
                context.move(to: CGPoint(x: xoffs + p1.x, y: yoffs - p1.y))
                context.addLine(to: CGPoint(x: xoffs + p2.x, y: yoffs - p2.y))
                // Seitenwand
                for i in 0 ... welt.areaLaenge
                {
                    let p1 = p3(x: 0, y: CGFloat (i), z: 0)
                    let p2 = p3(x: 0, y: CGFloat(i), z: CGFloat(welt.areaHoehe))
                    context.move(to: CGPoint(x: xoffs + p1.x, y: yoffs - p1.y))
                    context.addLine(to: CGPoint(x: xoffs + p2.x, y: yoffs - p2.y))
                    
                }
                p1 = p3(x: 0, y: CGFloat (welt.areaLaenge), z: CGFloat(welt.areaHoehe))
                p2 = p3(x: 0, y: 0, z: CGFloat(welt.areaHoehe))
                context.move(to: CGPoint(x: xoffs + p1.x, y: yoffs - p1.y))
                context.addLine(to: CGPoint(x: xoffs + p2.x, y: yoffs - p2.y))
                context.strokePath()

                // Windrose
                context.setLineDash(phase: 0, lengths: [])
                p1 = p3(x: -2, y: 0, z: CGFloat(welt.areaHoehe))
                p2 = p3(x: -2, y: 2, z: CGFloat(welt.areaHoehe))
                context.move(to: CGPoint(x: xoffs + p1.x, y: yoffs - p1.y))
                context.addLine(to: CGPoint(x: xoffs + p2.x, y: yoffs - p2.y))
                context.move(to: CGPoint(x: xoffs + p1.x-10.0, y: yoffs - (p1.y+5.0)))
                context.addLine(to: CGPoint(x: xoffs + p1.x, y: yoffs - p1.y))
                context.move(to: CGPoint(x: xoffs + p1.x-5.0, y: yoffs - (p1.y+10.0)))
                context.addLine(to: CGPoint(x: xoffs + p1.x, y: yoffs - p1.y))
                context.strokePath()
                context.textPosition = CGPoint(x: xoffs + p1.x+4, y: yoffs - (p1.y + asc))
                CTLineDraw (line, context)

                // sichtbare Ziegel, Quader und Marke zeichnen
                // wenn an der Stelle Karol steht, dann diesen zeichnen
                for a in 1 ... welt.areaBreite
                {
                    for b in 1 ... welt.areaLaenge
                    {
                    	//Achtung: c läuft bewusst höher als der Stapel, um die MArke richtig zu setzen
                        for c in 0 ... welt.areaStapelInhalt[a][b].count
                        {
                            let part = welt.getPart(x: a, y: b, z: c)
                            if part.was == welt.st_Ziegel
                            {
                                let p1 = p3(x: CGFloat(a-1), y: CGFloat(b), z: CGFloat(c))
                                let r = CGRect(x: xoffs + p1.x, y: yoffs-p1.y, width: ziegelBreite, height: ziegelHöhe)
                                switch part.farbe
                                {
                                    case BildFarben.rot:
                                        context.draw(ziegel_rot, in: r)
                                    case BildFarben.gelb:
                                        context.draw(ziegel_gelb, in: r)
                                    case BildFarben.grün:
                                        context.draw(ziegel_grün, in: r)
                                    case BildFarben.blau:
                                        context.draw(ziegel_blau, in: r)
                                    default:
                                        context.draw(ziegel_rot, in: r)
                                }
                            }
                            if part.was == welt.st_Quader
                            {
                                let p1 = p3(x: CGFloat(a-1), y: CGFloat(b), z: 0)
                                //p1.y = p1.y - quaderImg.getHeight()
                                let r = CGRect(x: xoffs + p1.x, y: yoffs-p1.y, width: quaderBreite, height: quaderHöhe)
                                context.draw(quader, in: r)
                            }
                            let marke = welt.getMarker(x: a, y: b)
                            if marke.ist && welt.brickCount(x: a, y: b) == c
                            {
                                let p1 = p3(x: CGFloat(a-1), y: CGFloat(b), z: CGFloat(c))
                                let r = CGRect(x: xoffs + p1.x, y: yoffs-p1.y, width: markenBreite, height: markenHöhe)
                                switch marke.farbe
                                {
                                    case BildFarben.rot:
                                        context.draw(marke_rot, in: r)
                                    case BildFarben.gelb:
                                        context.draw(marke_gelb, in: r)
                                    case BildFarben.grün:
                                        context.draw(marke_grün, in: r)
                                    case BildFarben.blau:
                                        context.draw(marke_blau, in: r)
                                    case BildFarben.schwarz:
                                        context.draw(marke_schwarz, in: r)
                                    default:
                                        context.draw(marke_gelb, in: r)
                                }
                            }
                            
                            for robo in welt.alleRoboter
                            {
                                if (a == robo.PositionXGeben()) && (b == robo.PositionYGeben()) && (c == welt.areaStapelInhalt[a][b].count)
                                        && (robo.SichtbarkeitGeben())
                                {
                                    let p1 = p3(x: CGFloat (a)-0.9, y: CGFloat (b)-0.3, z: CGFloat (c))
                                    let richtung = robo.getBlickrichtungNr()
                                    let knr = min(max(robo.KennungGeben()-1,0),Welt.maxRoboter-1)
                                    p1.x -=  CGFloat (roboter[richtung][knr].width - 30) / 2.0
                                    let r = CGRect(x: xoffs + p1.x, y: yoffs-p1.y, width: CGFloat (roboter[richtung][knr].width), height: CGFloat (roboter[richtung][knr].height))
                                    context.draw(roboter[richtung][knr], in: r)
                                }
                            }
                        }
                    }
                }
            }
        }
	}
    
    func WeltSetzen (wneu: Welt)
    {
    	welt = wneu
        WeltGroesseErmitteln()
    }
    
    func zeichneWelt ()
    {
        #if os(macOS)
        self.needsDisplay = true
        #elseif os(iOS)
        self.setNeedsDisplay()
        #endif
    }

    #if os(macOS)
	override func mouseDown(with theEvent: NSEvent)
    {
        Swift.print("MouseDown");
    }
    #endif
}


/**
Fenster zur Darstellung einer Karolwelt

Wird von Objekten der Klasse Welt direkt erzeugt.
Enthält eine Zeichenfläche zur Darstellung der Welt und
ein Textfeld zur Ausgabe von Fehlertexten.
- author: Ulli Freiberger
- version: 1.0 (31.12.2007)

Umsetzung der Klassen nach Swift für die Entwicklung mit Xcode.
- author: Albert Wiedemann
- version: 1.0
*/
class WeltFenster: View
{
	fileprivate var weltAnzeige: WeltAnzeige! = nil
    #if os(macOS)
    typealias TextField = NSTextField
    #elseif os(iOS)
    typealias TextField = UITextView
    #endif
    private var ausgabeText: TextField! = nil

	/*
    Standardkonstruktor
    */
	override init (frame: CGRect)
    {
    	super.init(frame: frame)
        ElementeAnlegen ()
    }

	/*
    Konstruktor für Erstellung aus NIB-File
    */
	required init? (coder: NSCoder)
    {
    	super.init (coder: coder)
        ElementeAnlegen ()
    }
    
    /**
    Legt die Oberflächenelemente an
    */
    private func ElementeAnlegen ()
    {
	    #if os(macOS)
        weltAnzeige = WeltAnzeige (frame: CGRect(x: 0.0, y: 80.0, width: frame.size.width, height: frame.size.height - 80.0))
        weltAnzeige.canDrawConcurrently = false
        addSubview(weltAnzeige)
        ausgabeText = TextField(frame: CGRect(x: 0.0, y: 0.0, width: frame.size.width, height: 80.0))
        fehlerAusgabeLoeschen()
        addSubview(ausgabeText)
        self.canDrawConcurrently=false
        #else
        weltAnzeige = WeltAnzeige (frame: CGRect(x: 0.0, y: 0.0, width: frame.size.width, height: frame.size.height - 80.0))
        addSubview(weltAnzeige)
        ausgabeText = TextField(frame: CGRect(x: 0.0, y: frame.size.height - 80.0, width: frame.size.width, height: 80.0))
        ausgabeText.isScrollEnabled=true
        ausgabeText.showsVerticalScrollIndicator=true
        ausgabeText.font = UIFont(name: "Helvatica", size: 18)
        fehlerAusgabeLoeschen()
        addSubview(ausgabeText)
        ausgabeText.setNeedsDisplay()
        #endif
        PlaygroundPage.current.liveView = self
    }

	/**
    Setzt die für diese Anzeige zu Grunde liegende Welt
    Muss nach der Initialisierung aufgerufen werden.
    Wird and die Welt-View weitergegeben.
    - parameters:
    	- w: Die zu setzende Welt
    */
    func WeltSetzen (w: Welt)
    {
    	weltAnzeige.WeltSetzen(wneu: w)
    }

	/**
    Löst das Neuzeichnen der Welt-View aus
    */
	func zeichnen()
	{
		weltAnzeige.zeichneWelt()
	}
	
    /**
     Löscht die ausgegebenen Fehlermeldungen.
    */
	func fehlerAusgabeLoeschen()
	{
		#if os(macOS)
        ausgabeText.stringValue = ""
        #else
        ausgabeText.text = ""
        #endif
	}
	
    /**
    Fügt den gegebenen Text an den Anzeigetext an.
    - parameters:
    	- text: der auszugebende Fehlertext
    */
	func fehlerAusgabe(text t: String)
	{
        #if os(macOS)
        ausgabeText.stringValue = ausgabeText.stringValue + t + "\n"
        #else
        ausgabeText.text = ausgabeText.text! + t + "\n"
        #endif
        //ausgabeText.scrollRangeToVisible(NSRange(ausgabeText.text.)
	}
}
